// lunzip - An unzipper with builtin LFN support
// Copyright (c) 2002, Sam Hawker <shawkie@geocities.com>
//
// This is free software under GPL. See the license.txt file for details.
//

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include <io.h>
#include <dos.h>
#include <dir.h>
#include <fcntl.h>
#include <SYS\stat.h>

#include "lfn.cpp"

#include "zlib.h"

#define BUFSIZE 16384UL

int mylog10(int x)
{
	int y = 0;
	while(x /= 10)
		y++;
	return y;
}

int myseekpath(const char *sdir, const char *lname, char *spath, char **psname, char *pcreated)
{
	DPB disk;
	longdirentry d, f;
	char drv, longpath[MAX_PATH_SIZE];

	drv = drive_of_path((char *)sdir);

	if(makeDPB(drv, disk) == 1)
		return -1;

	char far *ret = seekPath(&disk, (char *)sdir, longpath, d, 3);
	if(!ret)
		return -1;

	free(ret);

	direntry de;

	dirpointer dp;
	dp.cluster = d.start_cluster;
	dp.relsector = 0;
	dp.entry_no = 0;

	f = FindMask(&disk, (char *)lname, &dp, 1, &de);

	if(!f.name[0]) {
		if(!pcreated)
			return -1;

		lockDrive(drv);
		if(makeDirectory(&disk, &d, (char *)lname, &f)) {
			unlockDrive(drv);
			return -1;
		}
		unlockDrive(drv);

		*pcreated = 1;
	}

	if(sdir && sdir != spath)
		strcpy(spath, sdir);

	int length = strlen(spath);
	if(length != 0 && spath[length - 1] != '\\' && (length != 2 || spath[1] != ':')) {
		strcat(spath, "\\");
		length++;
	}

	char *ptr = spath + length;
	if(psname)
		*psname = ptr;

	int i;
	for(i = 0;(i < 8 && f.dosname[i] != ' '); i++)
		*ptr++ = f.dosname[i];
	if(f.dosname[8] != ' ')
		*ptr++ = '.';
	for(i = 8;(i < 11 && f.dosname[i] != ' '); i++)
		*ptr++ = f.dosname[i];
	*ptr = '\0';

	return 0;
}

int mycreatepath(const char *sdir, const char *lname, char *spath, char **psname)
{
	int i = 0, tilda, tildapos;
	char *ptr = (char *)lname, *dot = (char *)strrchr(lname, '.'), *nul = (char *)strchr(lname, '\0');

	char name[] = { 0, 0, 0, 0, 0, 0, 0, 0, 0 }, ext[] = { 0, 0, 0, 0 };

	if(dosStr((char *)lname) && ((!dot && (nul - lname) <= 8) || (dot && dot == strchr(lname, '.') && (dot - lname) <= 8 && (nul - dot) <= 4))) {
		for(i = 0; (i < 8 && *ptr && *ptr != '.'); ptr++)
			name[i++] = asciiUpcase[*ptr];
		tilda = 1;
		tildapos = min(i, 6);
		if(dot) {
			ptr = dot + 1;
			for(i = 0;(i < 3 && *ptr);)
				ext[i++] = asciiUpcase[*ptr++];
		}
	}
	else {
		for(i = 0; (i < 6 && *ptr && *ptr != '.'); ptr++) {
			if(dosChar(*ptr))
				name[i++] = asciiUpcase[*ptr];
			else if(winChar(*ptr))
				name[i++] = '_';
		}
		tilda = 2;
		tildapos = i;
		name[i++] = 126;
		name[i++] = '1';
		if(dot) {
			ptr = dot + 1;
			for(i = 0; (i < 3 && *ptr); ptr++) {
				if(dosChar(*ptr))
					ext[i++] = asciiUpcase[*ptr];
				else if(winChar(*ptr))
					ext[i++] = '_';
			}
		}
	}

	if(sdir && sdir != spath)
		strcpy(spath, sdir);

	int length = strlen(spath);
	if(length != 0 && spath[length - 1] != '\\' && (length != 2 || spath[1] != ':')) {
		if(length + 1 >= MAXPATH)
			return -1;
		strcat(spath, "\\");
		length++;
	}

	ptr = spath + length;
	if(psname)
		*psname = ptr;

	while(1) {
		if(length + strlen(name) + (dot ? 1 : 0) + strlen(ext) >= MAXPATH)
			return -1;

		strcpy(ptr, name);
		if(dot) {
			strcat(ptr, ".");
			strcat(ptr, ext);
		}

		ffblk ffblk;
		if(findfirst(spath, &ffblk, -1))
			break;

		if(tildapos + mylog10(tilda) > 6) {
			tildapos--;
			if(tildapos < 0)
				return -1;
		}

		name[tildapos] = 126;
		itoa(tilda, name + tildapos + 1, 10);

		tilda++;
	}

	return 0;
}

struct ren
{
	ren *rnext;
	char sname[13];
	char lname[1];
};

struct branch
{
	branch *head;
	branch *next;
	ren *rhead;
	char empty;
	char sname[13];
	char lname[1];
};

ren *newrename(branch *b, const char *lname, const char *sname)
{
	if(!strcmp(lname, sname))
		return 0;

	ren *rnew = (ren *)malloc(sizeof(ren) + strlen(lname));
	if(!rnew) {
		printf("Out of memory\n");
		return 0;
	}

	rnew->rnext = b->rhead;
	strcpy(rnew->sname, sname);
	strcpy(rnew->lname, lname);
	b->rhead = rnew;
	return rnew;
}

branch *newbranch(branch *b, const char *lname, const char *sname, char empty)
{
	branch *bnew = (branch *)malloc(sizeof(branch) + strlen(lname));
	if(!bnew) {
		printf("Out of memory\n");
		return 0;
	}

	bnew->head = 0;
	bnew->next = b->head;
	bnew->rhead = 0;
   bnew->empty = empty;
	strcpy(bnew->sname, sname);
	strcpy(bnew->lname, lname);
	b->head = bnew;
	return bnew;
}

void deletebranch(branch *b)
{
	ren *rcur = b->rhead;
	while(rcur) {
		ren *rnext = rcur->rnext;
		free(rcur);
		rcur = rnext;
	}
	branch *bcur = b->head;
	while(bcur) {
		branch *bnext = bcur->next;
		deletebranch(bcur);
		bcur = bnext;
	}
	free(b);
}

branch *lpathtospath(branch *b, const char *lpath, char *spath)
{
	char lname[MAX_PATH_SIZE];

	char *ptr = (char *)strchr(lpath, '\\');
	if(ptr) {
		strncpy(lname, lpath, ptr - lpath);
		lname[ptr - lpath] = '\0';
	}
	else
		strcpy(lname, lpath);

	branch *bcur = b->head;
	while(bcur) {
		if(!stricmp(bcur->lname, lname)) {
			strcat(spath, bcur->sname);
			strcat(spath, "\\");
			break;
		}
		bcur = bcur->next;
	}

	if(!bcur)
	{
		char *sname = 0;

		char created = 0;
		if(myseekpath(spath, lname, spath, &sname, &created)) {
			printf("Unable to create directory\n");
			return 0;
		}
		bcur = newbranch(b, lname, sname, created);
	}

	if(ptr)
		return lpathtospath(bcur, ptr + 1, spath);
	else
		return bcur;
}

void dorenames(branch *b, char *spath)
{
	char *ptr = spath + strlen(spath);

	if(b->sname[0]) {
		strcpy(ptr, b->sname);
		strcat(ptr, "\\");
	}

	{
		DPB disk;
		direntry de;
		dirpointer dp;
		longdirentry d, f, n;
		char drv, longpath[MAX_PATH_SIZE];

		drv = drive_of_path((char *)spath);

		if(makeDPB(drv, disk) == 1)
			return;

		char far *ret = seekPath(&disk, (char *)spath, longpath, d, 3);
		if(!ret)
			return;

		free(ret);

		lockDrive(drv);

		ren *rcur = b->rhead;
		while(rcur) {
			dp.cluster = d.start_cluster;
			dp.relsector = 0;
			dp.entry_no = 0;

			n = f = FindMask(&disk, (char *)rcur->sname, &dp, 6, &de);

			int outlen;
			xlate_to_uni((unsigned char *)rcur->lname, strlen(rcur->lname), (unsigned char *)n.uniname, &outlen, ':');
			n.uniname[outlen] = 0;

			dp.entry_no--;
			deleteDirentry(&disk, &d, &dp, &f);

			if(insertDirentry(&disk, &d, &n, PROTECT_ALL))
				insertDirentry(&disk, &d, &f, PROTECT_ALL);

			rcur = rcur->rnext;
		}

		unlockDrive(drv);
	}

	branch *bcur = b->head;
	while(bcur) {
		dorenames(bcur, spath);
		bcur = bcur->next;
	}

	*ptr = '\0';
}

struct eocd
{
	word wThisDisk;
	word wDisk;
	word wThisCount;
	word wCount;
	dword dwSize;
	dword dwOffset;
};

struct fh
{
	dword dwSig;
	word wVerMade;
	word wVerExtract;
	word wBitFlag;
	word wComprMethod;
	word wFileTime;
	word wFileDate;
	dword dwCRC;
	dword dwComprSize;
	dword dwUncomprSize;
	word wFileNameLen;
	word wExtraFieldLen;
	word wCommentLen;
	word wDisk;
	word wIntAttr;
	dword dwAttr;
	dword dwOffset;
};

struct lfh
{
	dword dwSig;
	word wVerExtract;
	word wBitFlag;
	word wComprMethod;
	word wFileTime;
	word wFileDate;
	dword dwCRC;
	dword dwComprSize;
	dword dwUncomprSize;
	word wFileNameLen;
	word wExtraFieldLen;
};

eocd myeocd;

word disk;

char *filename = 0;

byte *mycd = 0;

FILE *askfordisk(FILE *fi, word disknum)
{
	fclose(fi);

	disk = disknum;
	printf("Please insert disk number %i and press a key to continue\n", disk + 1);
	getch();

	fi = fopen(filename, "rb");

	if(!fi)
		printf("Unable to open zip file\n");

	return fi;
}

int readcentraldir(FILE *fp)
{
	dword dwSignature;
	for(word wTestLen = 0; wTestLen < 0xFFFF; wTestLen++) {
		fseek(fp, -(2L + wTestLen), SEEK_END);

		word wCommentLen;
		fread(&wCommentLen, 2, 1, fp);

		if(wCommentLen == wTestLen) {
			fseek(fp, -(22L + wTestLen), SEEK_CUR);

			fread(&dwSignature, 4, 1, fp);

			if(dwSignature == 0x06054b50)
				break;
		}
	}

	if(dwSignature != 0x06054b50)
		return -1;

	fread(&myeocd, sizeof(eocd), 1, fp);

	disk = myeocd.wThisDisk;

	if(myeocd.wDisk != disk) {
		if((fp = askfordisk(fp, myeocd.wDisk)) == 0)
			return -1;
	}

	mycd = (byte *)farmalloc(myeocd.dwSize);
	if(!mycd) {
		free(mycd);
		printf("Out of memory");
		return -1;
	}

	fseek(fp, myeocd.dwOffset, SEEK_SET);

	byte huge *pos = mycd;

	dword total = myeocd.dwSize;

	while(total) {
		word current = min(total, BUFSIZE);
		word actual = fread(pos, 1, current, fp);
		total -= actual;
		pos += actual;
		if(actual != current) {
			disk++;
			printf("Please insert disk number %i and press a key to continue\n", disk + 1);
			getch();
		}
	}

	return 0;
}

void extractfile(FILE *fi, FILE *fo, fh *pfh)
{
	byte *bufExtract = (byte *)malloc(BUFSIZE);
	if(!bufExtract) {
		printf("Out of memory\n");
		return;
	}

	dword crc = crc32(0, Z_NULL, 0);

	dword total = pfh->dwComprSize;

	printf("Extracting");

	while(total) {
		word current = min(BUFSIZE, total);
		word actual = fread(bufExtract, 1, current, fi);
		printf(".");
		if(actual) {
			fwrite(bufExtract, 1, actual, fo);
			crc = crc32(crc, bufExtract, actual);
			total -= actual;
		}
		if(actual != current) {
			printf("\n");
			if((fi = askfordisk(fi, disk + 1)) == 0)
				return;
		}
	}

	printf("\n");

	if(crc != pfh->dwCRC)
		printf("File failed CRC32 check\n");

	free(bufExtract);
}

void inflatefile(FILE *fi, FILE *fo, fh *pfh)
{
	byte *bufCompr = (byte *)malloc(BUFSIZE);
	byte *bufUncompr = (byte *)malloc(BUFSIZE);
	if(!bufCompr || !bufUncompr) {
		printf("Out of memory\n");
		free(bufCompr);
		free(bufUncompr);
	}

	int err = Z_OK;

	z_stream stream;

	stream.zalloc=(alloc_func)0;
	stream.zfree=(free_func)0;
	stream.opaque=(voidpf)0;

	inflateInit2(&stream, -MAX_WBITS);

	dword crc = crc32(0, Z_NULL, 0);

	dword total = pfh->dwComprSize;

	printf("Inflating");

	while(total) {
		word current = min(BUFSIZE, total);
		word actual = fread(bufCompr, 1, current, fi);
		printf(".");
		if(actual) {
			stream.next_in = bufCompr;
			stream.avail_in = actual;
			do {
				stream.next_out = bufUncompr;
				stream.avail_out = BUFSIZE;
				err = inflate(&stream, Z_SYNC_FLUSH);
				fwrite(bufUncompr, 1, BUFSIZE - stream.avail_out, fo);
				crc = crc32(crc, bufUncompr, BUFSIZE - stream.avail_out);
			} while(err == Z_OK && !stream.avail_out);
			total -= actual;
		}
		if(actual != current) {
			printf("\n");
			if((fi = askfordisk(fi, disk + 1)) == 0)
				return;
		}
	}

	printf("\n");

	if(err == Z_DATA_ERROR)
		printf("Invalid compressed data to inflate\n");

	if(crc != pfh->dwCRC)
		printf("File failed CRC32 check\n");

	inflateEnd(&stream);

	free(bufCompr);
	free(bufUncompr);
}

void main(int argc, char *argv[])
{
	char update = 0, replace = 0, drives = 0;

	for(int j = 1; j < argc; j++) {
		if(argv[j][0] == '/') {
			char *ptr = argv[j] + 1;
			while(*ptr) {
				switch(*ptr) {
				case 'U':
				case 'u':
					update = 1;
					break;
				case 'R':
				case 'r':
					replace = 1;
					break;
				case 'D':
				case 'd':
					drives = 1;
					break;
				default:
					printf("Invalid switch /%c\n", *ptr);
				}
				ptr++;
			}
		}
		else {
			if(filename)
				printf("Too many arguments\n");
			else
				filename = argv[j];
		}
	}

	if(!filename) {
		printf("lunzip - An unziper with builtin LFN support\n");
		printf("Copyright (c) 2002, Sam Hawker <shawkie@geocities.com>\n\n");
		printf("This is free software under GPL. See the license.txt file for details.\n\n");
		printf("Syntax:\n\n   lunzip [/U|R] [/D] [filename.zip]\n\n");
		printf("Switches:\n\n");
		printf("   /U   Update  - replace older files\n");
		printf("   /R   Replace - replace all files\n");
		printf("   /D   Drives  - convert \"Drive X\\\" to \"X:\\\" in paths\n");
		return;
	}

	FILE *fi = fopen(filename, "rb");

	if(!fi) {
		printf("Unable to open zip file\n");
		return;
	}

	if(readcentraldir(fi)) {
		printf("Invalid zip file or not last disk\n");
		fclose(fi);
		return;
	}

	branch *mybranch = (branch *)malloc(sizeof(branch));
	mybranch->head = 0;
	mybranch->next = 0;
	mybranch->rhead = 0;
	mybranch->sname[0] = '\0';
	mybranch->lname[0] = '\0';

	byte huge *pos = 0;

	fh myfh;
	lfh mylfh;

	char spath[MAXPATH];
	char lpath[MAX_PATH_SIZE];

	for(word i = 0; i < myeocd.wCount; i++) {

		if(pos)
			pos += sizeof(fh) + myfh.wFileNameLen + myfh.wExtraFieldLen + myfh.wCommentLen;
		else
			pos = mycd;

		_fmemcpy(&myfh, pos, sizeof(fh));

		if(myfh.dwSig != 0x02014b50) {
			printf("Invalid zip file\n");
			break;
		}

		if(myfh.wFileNameLen >= MAX_PATH_SIZE) {
			printf("Filename too long\n");
			break;
		}

		_fmemcpy(lpath, pos + sizeof(fh), myfh.wFileNameLen);
		lpath[myfh.wFileNameLen] = '\0';

		spath[0] = '\0';

		branch *b = mybranch;

		char *ptr = lpath, *lname = lpath, *sname = 0;
		while((ptr = strchr(ptr, '/')) != 0)
			*ptr++ = '\\';

		if(drives && !strncmp(lpath, "Drive ", 6) && lpath[6] >= 'A' && lpath[6] <= 'Z' && lpath[7] == '\\') {
			memmove(lpath, lpath + 5, MAX_PATH_SIZE - 5);
			lpath[0] = lpath[1];
			lpath[1] = ':';
		}

		ptr = strrchr(lpath, '\0');
		if(ptr != lpath && *(ptr-1) == '\\')
			*(ptr-1) = '\0';

		printf("%s\n", lpath);

		if(myfh.dwAttr & 0x10) {
			if(!lpathtospath(mybranch, lpath, spath)) {
				printf("Unable to create directory\n");
				continue;
			}
			ptr = strrchr(spath, '\0');
			if(ptr != spath && *(ptr-1) == '\\')
				*(ptr-1) = '\0';
			if(_dos_setfileattr(spath, myfh.dwAttr & ~_A_SUBDIR))
				printf("Unable to set attributes\n");
			continue;
		}

		ptr = strrchr(lpath, '\\');
		if(ptr) {
			*ptr = '\0';
			lname = ptr + 1;
			if((b = lpathtospath(mybranch, lpath, spath)) == 0) {
				printf("Unable to create directory\n");
				continue;
			}
			*ptr = '\\';
		}

		if(b->empty || myseekpath(0, lname, spath, 0, 0)) {
			if(mycreatepath(0, lname, spath, &sname)) {
				printf("Unable to create file\n");
				continue;
			}
		} else {
			if(update) {
				ffblk ffblk;
				findfirst(spath, &ffblk, -1);
				if(ffblk.ff_fdate < myfh.wFileDate || (ffblk.ff_fdate == myfh.wFileDate && ffblk.ff_ftime <= myfh.wFileTime)) {
					printf("File already exists\n");
					continue;
				}
			}
			else if(!replace) {
				printf("File already exists\n");
				continue;
			}
			_dos_setfileattr(spath, _A_NORMAL);
		}

		if(myfh.wBitFlag & 1) {
			printf("Encryption not supported\n");
			continue;
		}

		if(myfh.wComprMethod && myfh.wComprMethod != 8) {
			printf("Unsupported compression method\n");
			continue;
		}

		if(myfh.wDisk != disk) {
			if((fi = askfordisk(fi, myfh.wDisk)) == 0)
				break;
		}

		fseek(fi, myfh.dwOffset, SEEK_SET);

		fread(&mylfh, sizeof(lfh), 1, fi);

		if(mylfh.dwSig != 0x04034b50) {
			printf("Unable to find local file header\n");
			continue;
		}

		fseek(fi, myfh.wFileNameLen + myfh.wExtraFieldLen, SEEK_CUR);

		FILE *fo = fopen(spath, "wb");

		if(!fo) {
			printf("Unable to create file\n");
			continue;
		}

		if(myfh.wComprMethod)
			inflatefile(fi, fo, &myfh);
		else
			extractfile(fi, fo, &myfh);

		if(_dos_setftime(fileno(fo), myfh.wFileDate, myfh.wFileTime))
			printf("Unable to set file date and time\n");

		fclose(fo);

		if(_dos_setfileattr(spath, myfh.dwAttr))
			printf("Unable to set attributes\n");

		if(sname)
			newrename(b, lname, sname);

		if(!fi)
			break;
	}

	free(mycd);

	spath[0] = '\0';

	printf("Renaming files...\n");

	dorenames(mybranch, spath);

	deletebranch(mybranch);

	fclose(fi);
}

